import argparse

import torch

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype, generate_initial_opr_mode_weights

import test_utils as tutils


class Test2dPtychoAutodiff(tutils.TungstenDataTester):

    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_autodiff')
    def test_2d_ptycho_autodiff(self):
        self.setup_ptychi(cpu_only=False)
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        
        options = api.AutodiffPtychographyOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 0.1
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 32
        
        task = PtychographyTask(options)
        task.run()
        
        recon = task.get_data_to_cpu('object', as_numpy=True)[0]
        return recon
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_autodiff_pos_corr')
    def test_2d_ptycho_autodiff_pos_corr(self):
        self.setup_ptychi(cpu_only=False)
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0, pos_type="nominal")
        
        options = api.AutodiffPtychographyOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 0.1
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = True
        options.probe_position_options.step_size = 10
        options.probe_position_options.correction_options.update_magnitude_limit = None
        options.probe_position_options.correction_options.clip_update_magnitude_by_mad = False
        options.probe_position_options.correction_options.differentiation_method = api.ImageGradientMethods.GAUSSIAN
        
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 32
        
        task = PtychographyTask(options)
        task.run()
        
        recon = task.get_data_to_cpu('object', as_numpy=True)[0]
        pos_corr = task.get_data_to_cpu('probe_positions', as_numpy=True)
        
        if self.debug and not self.generate_gold:
            _, _, _, pos_true = self.load_tungsten_data(additional_opr_modes=0, pos_type="true")
            tutils.plot_probe_positions(pos_corr, pos_true)
        
        return recon
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_autodiff_l1')
    def test_2d_ptycho_autodiff_l1(self):
        self.setup_ptychi(cpu_only=False)
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)
        
        options = api.AutodiffPtychographyOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 0.1
        options.object_options.l1_norm_constraint.enabled = True
        options.object_options.l1_norm_constraint.weight = 1e-3
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 32
        
        task = PtychographyTask(options)
        task.run()
        
        recon = task.get_data_to_cpu('object', as_numpy=True)[0]
        return recon
            
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_autodiff_opr')
    def test_2d_ptycho_autodiff_opr(self):
        
        self.setup_ptychi(cpu_only=False)
        
        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=3)
        
        options = api.AutodiffPtychographyOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 0.1
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 0.1

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        
        options.opr_mode_weight_options.initial_weights = generate_initial_opr_mode_weights(len(positions_px), probe.shape[0])
        options.opr_mode_weight_options.optimizable = True
        options.opr_mode_weight_options.optimizer = api.Optimizers.ADAM
        options.opr_mode_weight_options.step_size = 1e-3
        
        options.reconstructor_options.batch_size = 96
        options.reconstructor_options.num_epochs = 32
        
        task = PtychographyTask(options)
        task.run()
        
        recon = task.get_data_to_cpu('object', as_numpy=True)[0]
        return recon
    
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()

    tester = Test2dPtychoAutodiff()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_autodiff()
    tester.test_2d_ptycho_autodiff_pos_corr()
    tester.test_2d_ptycho_autodiff_l1()
    tester.test_2d_ptycho_autodiff_opr()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import numpy as np
from typing import Optional
import h5py
from PIL import Image, ImageDraw, ImageFont
import cv2
import os
from .settings import MovieFileTypes


dataset_name = "frames"


def append_array_to_h5(array: np.ndarray, file_path: str, create_new_file: bool = False):
    if not os.path.exists(file_path) or create_new_file:
        # Create new file and 3D dataset
        with h5py.File(file_path, "w") as f:
            max_shape = (
                None,
                array.shape[0],
                array.shape[1],
            )  # Unlimited depth, fixed height & width
            dset = f.create_dataset(
                dataset_name,
                shape=(1, array.shape[0], array.shape[1]),  # Initial shape (depth=1)
                maxshape=max_shape,
                dtype=array.dtype,
            )
            dset[0] = array  # Store the first array
    else:
        # Append data to the existing 3D dataset
        with h5py.File(file_path, "a") as f:
            dset = f[dataset_name]
            new_depth = dset.shape[0] + 1  # Increase depth by 1
            dset.resize((new_depth, array.shape[0], array.shape[1]))  # Expand along depth
            dset[-1] = array  # Append new 2D array at the last depth


def save_movie_to_file(
    array: np.ndarray,
    file_type: MovieFileTypes,
    output_path: str,
    fps: int = 30,
    colormap: int = cv2.COLORMAP_BONE,
    enhance_contrast: bool = False,
    titles: list[str] = None,
    compress: bool = True,
    upper_bound: Optional[float] = None,
    lower_bound: Optional[float] = None,
):
    if upper_bound is not None:
        array = np.clip(array, a_max=upper_bound, a_min=None)
    if lower_bound is not None:
        array = np.clip(array, a_min=lower_bound, a_max=None)

    if file_type == MovieFileTypes.GIF:
        numpy_to_gif(array, output_path, fps, colormap, enhance_contrast, titles, compress=compress)
    elif file_type == MovieFileTypes.MP4:
        numpy_to_mp4(array, output_path, fps, colormap, enhance_contrast)


def numpy_to_gif(
    array: np.ndarray,
    output_path: str,
    fps: int = 30,
    colormap: int = cv2.COLORMAP_BONE,
    enhance_contrast: bool = False,
    titles: list[str] = None,
    font_size: int = 20,
    compress: bool = True,
):
    """
    Convert a 3D NumPy array (frames along the first axis) to a GIF file.

    Parameters:
        array (np.ndarray): 3D NumPy array of shape (frames, height, width)
        output_path (str): Path to save the GIF file.
        fps (int): Frames per second for the output video.
        colormap (int): OpenCV colormap to apply.
        enhance_contrast (bool): Whether to apply contrast enhancement (default: False).
    """
    frames, height, width = array.shape[:3]
    is_grayscale = len(array.shape) == 3
    gif_frames = []

    if titles and len(titles) != frames:
        raise ValueError("Length of titles list must match the number of frames.")
    
    # Apply contrast enhancement if needed
    if enhance_contrast:
        array = np.power(array, 0.5)  # Apply gamma correction to boost small values

    # Normalize to [0, 255] and convert to uint8
    array = cv2.normalize(array, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)

    for i in range(frames):
        frame = array[i]

        if is_grayscale:
            frame = cv2.applyColorMap(frame, colormap)  # Apply colormap to grayscale frame

        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)  # Convert to RGB for PIL

        pil_frame = Image.fromarray(frame)

        # Draw title if provided
        if titles:
            draw = ImageDraw.Draw(pil_frame)
            try:
                font = ImageFont.truetype("arial.ttf", font_size)
            except IOError:
                font = ImageFont.load_default()

            text = titles[i]
            text_bbox = draw.textbbox((0, 0), text, font=font)  # Get text bounding box
            text_width = text_bbox[2] - text_bbox[0]  # Width of the text
            # text_height = text_bbox[3] - text_bbox[1]  # Height of the text
            position = ((width - text_width) // 2, 10)  # Centered at the top
            draw.text(position, text, font=font, fill=(255, 255, 255))

        gif_frames.append(pil_frame)

    # Save frames as GIF
    gif_frames[0].save(
        output_path,
        save_all=True,
        append_images=gif_frames[1:],
        duration=int(1000 / fps),
        loop=0,
        optimize=compress,
    )
    print(f"GIF saved to {output_path}")


def numpy_to_mp4(
    array: np.ndarray,
    output_path: str,
    fps: int = 30,
    colormap: int = cv2.COLORMAP_BONE,
    enhance_contrast: bool = False,
):
    """
    Convert a 3D NumPy array (frames along the first axis) to an MP4 file.

    Parameters:
        array (np.ndarray): 3D NumPy array of shape (frames, height, width).
        output_path (str): Path to save the MP4 file.
        fps (int): Frames per second for the output video.
        colormap (int): OpenCV colormap to apply.
        enhance_contrast (bool): Whether to apply contrast enhancement (default: False).
    """
    frames, height, width = array.shape[:3]
    is_grayscale = len(array.shape) == 3

    fourcc = cv2.VideoWriter_fourcc(*"mp4v")  # Codec for MP4
    video_writer = cv2.VideoWriter(output_path, fourcc, fps, (width, height))

    for i in range(frames):
        frame = array[i]

        # Apply contrast enhancement if needed
        if enhance_contrast:
            frame = np.power(frame, 0.5)  # Apply gamma correction to boost small values

        # Normalize to [0, 255] and convert to uint8
        frame = cv2.normalize(frame, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)

        if is_grayscale:
            frame = cv2.applyColorMap(frame, colormap)  # Apply colormap to grayscale frame

        video_writer.write(frame)

    video_writer.release()
    print(f"Video saved to {output_path}")



import argparse
import logging

import torch

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype, generate_initial_opr_mode_weights

import test_utils as tutils


class Test2dPtychoLsqmlMultiscan(tutils.TungstenDataTester):
    
    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_lsqml_multiscan')
    def test_2d_ptycho_lsqml_multiscan(self):        
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(pos_type='true')
        
        # Split data and positions
        data1 = data[:500]
        data2 = data[500:]
        positions_px_1 = positions_px[:500]
        positions_px_2 = positions_px[500:]
        
        # Create task 1
        options_1 = api.LSQMLOptions()
        options_1.data_options.data = data1
        
        options_1.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options_1.object_options.pixel_size_m = pixel_size_m
        options_1.object_options.optimizable = True
        options_1.object_options.optimizer = api.Optimizers.SGD
        options_1.object_options.step_size = 1
        options_1.object_options.build_preconditioner_with_all_modes = True
        
        options_1.probe_options.initial_guess = probe
        options_1.probe_options.optimizable = True
        options_1.probe_options.optimizer = api.Optimizers.SGD
        options_1.probe_options.step_size = 1

        options_1.probe_position_options.position_x_px = positions_px_1[:, 1]
        options_1.probe_position_options.position_y_px = positions_px_1[:, 0]
        options_1.probe_position_options.optimizable = False
        
        options_1.reconstructor_options.batch_size = 96
        options_1.reconstructor_options.noise_model = api.NoiseModels.GAUSSIAN
        options_1.reconstructor_options.num_epochs = 8
        options_1.reconstructor_options.allow_nondeterministic_algorithms = False
        
        task_1 = PtychographyTask(options_1)
        
        # Create task 2
        options_2 = api.LSQMLOptions()
        options_2.data_options.data = data2
        
        options_2.object_options.initial_guess = torch.ones([1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], dtype=get_default_complex_dtype())
        options_2.object_options.pixel_size_m = pixel_size_m
        options_2.object_options.optimizable = True
        options_2.object_options.optimizer = api.Optimizers.SGD
        options_2.object_options.step_size = 1
        options_2.object_options.build_preconditioner_with_all_modes = True
        
        options_2.probe_options.initial_guess = probe
        options_2.probe_options.optimizable = True
        options_2.probe_options.optimizer = api.Optimizers.SGD
        options_2.probe_options.step_size = 1

        options_2.probe_position_options.position_x_px = positions_px_2[:, 1]
        options_2.probe_position_options.position_y_px = positions_px_2[:, 0]
        options_2.probe_position_options.optimizable = False
        
        options_2.reconstructor_options.batch_size = 96
        options_2.reconstructor_options.noise_model = api.NoiseModels.GAUSSIAN
        options_2.reconstructor_options.num_epochs = 8
        options_2.reconstructor_options.allow_nondeterministic_algorithms = False
        
        task_2 = PtychographyTask(options_2)
        
        # Disable progress bar for task 2
        task_2.reconstructor.pbar.disable = True
        
        # Run tasks each for one epoch each time
        all_tasks = [task_1, task_2]
        for i_epoch in range(options_1.reconstructor_options.num_epochs):
            for i_task, task in enumerate(all_tasks):
                task.run(1)
                # Copy object to next task
                if i_task < len(all_tasks) - 1:
                    all_tasks[i_task + 1].copy_data_from_task(task, params_to_copy=("object",))
        
        recon = all_tasks[-1].get_data_to_cpu('object', as_numpy=True)[0]
        return recon
    
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()

    tester = Test2dPtychoLsqmlMultiscan()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_lsqml_multiscan()
